#include "render_view.h"
#include "renderer_api.h"

#include "runtime/config.h"
#include "runtime/mem_arena.h"
#include "runtime/threading.h"

RT_CVAR_I(rt_RenderViewArenaSize,
          "Size of the memory arena used for allocating render views. Default: 1 MB",
          RT_MB(1));

static rt_arena _view_arena;
static rt_mutex *_view_lock;

rt_result InitRenderViews(void) {
    rt_create_arena_result arena_res = rtCreateArena(NULL, (size_t)rt_RenderViewArenaSize.i);
    if (!arena_res.ok)
        return RT_OUT_OF_MEMORY;
    _view_arena = arena_res.arena;
    _view_lock  = rtCreateMutex();
    if (!_view_lock) {
        rtReleaseArena(&_view_arena);
        return RT_UNKNOWN_ERROR;
    }
    return RT_SUCCESS;
}

void ShutdownRenderViews(void) {
    rtDestroyMutex(_view_lock);
    rtReleaseArena(&_view_arena);
}

void ResetRenderViews(void) {
    rtArenaClear(&_view_arena);
}

RT_DLLEXPORT rt_create_render_view_result rtCreateRenderView(const rt_render_object_type *types,
                                                             uint32_t type_count) {

#ifdef RT_DEBUG
    for (uint32_t i = 0; i < type_count - 1; ++i) {
        for (uint32_t j = i + 1; j < type_count; ++j) {
            RT_ASSERT(types[i] != types[j], "Duplicate render list type detected.");
        }
    }
#endif

    size_t size = type_count * (sizeof(rt_render_list) + sizeof(rt_render_list));
    rtLockMutex(_view_lock);
    void *storage = rtArenaPush(&_view_arena, size);
    rtUnlockMutex(_view_lock);
    if (!storage) {
        return (rt_create_render_view_result){
            .ok = false,
        };
    }

    rt_render_view view;
    view.lists      = storage;
    view.list_types = (rt_render_object_type *)(view.lists + type_count);
    view.list_count = type_count;
    return (rt_create_render_view_result){.ok = true, .view = view};
}

RT_DLLEXPORT bool
rtPushRenderObjectToView(rt_render_view *view, rt_render_object_type type, const void *object) {
    for (uint32_t i = 0; i < view->list_count; ++i) {
        if (view->list_types[i] == type)
            return rtPushRenderListEntry(&view->lists[i], object);
    }
    return false;
}

RT_DLLEXPORT void
rtSubmitRenderView(rt_render_view view, rt_render_graph *render_graph, uint32_t pass_id) {
    g_renderer.SubmitRenderView(render_graph, pass_id, view);
}